#include "pch.h"
#include "BasicTextRenderer.h"

using namespace D2D1;
using namespace Microsoft::WRL;

// Constructor
BasicTextRenderer::BasicTextRenderer() :
    m_refCount(0)
{
}

// IUnknown methods
ULONG STDMETHODCALLTYPE BasicTextRenderer::AddRef()
{
    m_refCount++;
    return m_refCount;
}

ULONG STDMETHODCALLTYPE BasicTextRenderer::Release()
{
    m_refCount--;
    LONG newCount = m_refCount;

    if (m_refCount == 0)
        delete this;

    return newCount;
}

HRESULT STDMETHODCALLTYPE BasicTextRenderer::QueryInterface(_In_ REFIID riid, 
                                                            _Outptr_ void** ppOutput)
{
    *ppOutput = nullptr;
    HRESULT hr = S_OK;

    if (riid == __uuidof(IDWriteTextRenderer))
    {
        *ppOutput = static_cast<IDWriteTextRenderer*>(this);
        AddRef();
    }
    else if (riid == __uuidof(IDWritePixelSnapping))
    {
        *ppOutput = static_cast<IDWritePixelSnapping*>(this);
        AddRef();
    }
    else if (riid == __uuidof(IUnknown))
    {
        *ppOutput = this;
        AddRef();
    }
    else
    {
        hr = E_NOINTERFACE;
    }
    return hr;
}

// IDWritePixelSnapping methods
HRESULT BasicTextRenderer::IsPixelSnappingDisabled(void * clientDrawingContext, 
                                                   _Out_ BOOL * isDisabled)
{
    *isDisabled = false;
    return S_OK;
}

HRESULT BasicTextRenderer::GetPixelsPerDip(void * clientDrawingContext,
                                           _Out_ FLOAT * pixelsPerDip)
{
    DrawingContext * drawingContext =
        static_cast<DrawingContext *>(clientDrawingContext);

    float dpiX, dpiY;
    drawingContext->renderTarget->GetDpi(&dpiX, &dpiY);
    *pixelsPerDip = dpiX / 96;
    return S_OK;
}

HRESULT BasicTextRenderer::GetCurrentTransform(void * clientDrawingContext,
                                               DWRITE_MATRIX * transform)
{
    DrawingContext * drawingContext = 
        static_cast<DrawingContext *>(clientDrawingContext);

    // Matrix structures are defined identically
    drawingContext->renderTarget->GetTransform((D2D1_MATRIX_3X2_F *) transform);
    return S_OK;
}

// IDWriteTextRenderer methods
HRESULT BasicTextRenderer::DrawGlyphRun(void * clientDrawingContext,
                                        FLOAT baselineOriginX,
                                        FLOAT baselineOriginY,
                                        DWRITE_MEASURING_MODE measuringMode,
                                        _In_ const DWRITE_GLYPH_RUN * glyphRun,
                                        _In_ const DWRITE_GLYPH_RUN_DESCRIPTION * 
                                            glyphRunDescription,
                                        IUnknown * clientDrawingEffect)
{
    DrawingContext * drawingContext =
        static_cast<DrawingContext *>(clientDrawingContext);

    ID2D1RenderTarget * renderTarget = drawingContext->renderTarget;

    // Get brush
    ID2D1Brush * brush = drawingContext->defaultBrush;
        
    if (clientDrawingEffect != nullptr)
    {
        brush = static_cast<ID2D1Brush *>(clientDrawingEffect);
    }

    renderTarget->DrawGlyphRun(Point2F(baselineOriginX, baselineOriginY), 
                               glyphRun, brush, measuringMode);
    return S_OK;
}

HRESULT BasicTextRenderer::DrawUnderline(void * clientDrawingContext,
                                         FLOAT baselineOriginX,
                                         FLOAT baselineOriginY,
                                         _In_ const DWRITE_UNDERLINE * 
                                             underline,
                                         IUnknown * clientDrawingEffect)

{
    FillRectangle(clientDrawingContext,
                  clientDrawingEffect,
                  baselineOriginX,
                  baselineOriginY + underline->offset,
                  underline->width,
                  underline->thickness,
                  underline->readingDirection,
                  underline->flowDirection);
    return S_OK;
}

HRESULT BasicTextRenderer::DrawStrikethrough(void * clientDrawingContext,
                                             FLOAT baselineOriginX,
                                             FLOAT baselineOriginY,
                                             _In_ const DWRITE_STRIKETHROUGH * 
                                                 strikethrough,
                                             IUnknown * clientDrawingEffect)
{
    FillRectangle(clientDrawingContext,
                  clientDrawingEffect,
                  baselineOriginX,
                  baselineOriginY + strikethrough->offset,
                  strikethrough->width,
                  strikethrough->thickness,
                  strikethrough->readingDirection,
                  strikethrough->flowDirection);
    return S_OK;
}

void BasicTextRenderer::FillRectangle(void * clientDrawingContext, 
                                      IUnknown * clientDrawingEffect,
                                      float x, float y, 
                                      float width, float thickness, 
                                      DWRITE_READING_DIRECTION readingDirection, 
                                      DWRITE_FLOW_DIRECTION flowDirection)
{
    DrawingContext * drawingContext =
        static_cast<DrawingContext *>(clientDrawingContext);

    // Get brush
    ID2D1Brush * brush = drawingContext->defaultBrush;

    if (clientDrawingEffect != nullptr)
    {
        brush = static_cast<ID2D1Brush *>(clientDrawingEffect);
    }

    D2D1_RECT_F rect = RectF(x, y, x + width, y + thickness);
    drawingContext->renderTarget->FillRectangle(&rect, brush);
}

HRESULT BasicTextRenderer::DrawInlineObject(void * clientDrawingContext,
                                            FLOAT originX,
                                            FLOAT originY,
                                            IDWriteInlineObject * inlineObject,
                                            BOOL isSideways,
                                            BOOL isRightToLeft,
                                            IUnknown * clientDrawingEffect)
{
    DrawingContext * drawingContext =
        static_cast<DrawingContext *>(clientDrawingContext);

    return inlineObject->Draw(clientDrawingContext, 
                              this, 
                              originX, 
                              originY, 
                              isSideways, 
                              isRightToLeft, 
                              clientDrawingEffect);
}
