﻿#include "pch.h"
#include "ColorTextFormattingDemoRenderer.h"
#include "Common/DirectXHelper.h"

using namespace ColorTextFormattingDemo;

using namespace D2D1;
using namespace Microsoft::WRL;

ColorTextFormattingDemoRenderer::ColorTextFormattingDemoRenderer(const std::shared_ptr<DX::DeviceResources>& deviceResources) :
    m_deviceResources(deviceResources),
    m_useCustomRenderer(false)
{
    // Create device independent resources
    DX::ThrowIfFailed(
        m_deviceResources->GetD2DFactory()->CreateDrawingStateBlock(&m_stateBlock)
        );

    ComPtr<IDWriteTextFormat> textFormat;

    DX::ThrowIfFailed(
        m_deviceResources->GetDWriteFactory()->CreateTextFormat(
            L"Times New Roman",
            nullptr,
            DWRITE_FONT_WEIGHT_NORMAL,
            DWRITE_FONT_STYLE_NORMAL,
            DWRITE_FONT_STRETCH_NORMAL,
            24.0f,
            L"en-us",
            &textFormat)
        );
    
    m_text = L"This paragraph of text rendered with "
             L"DirectWrite is based on "
             L"IDWriteTextFormat and IDWriteTextLayout "
             L"objects and is capable of different "
             L"RBG RGB foreground colors, such as red, "
             L"green, and blue by passing brushes to "
             L"the SetDrawingEffect method.";

    DX::ThrowIfFailed(
        m_deviceResources->GetDWriteFactory()->CreateTextLayout(
            m_text.c_str(),
            (uint32) m_text.length(),
            textFormat.Get(),
            440.0f,         // Max width for paragraph
            std::numeric_limits<float>::infinity(),
            &m_textLayout)
        );

    // Set character formatting
    DWRITE_TEXT_RANGE textRange;

    std::wstring strFind = L"IDWriteTextFormat";
    textRange.startPosition = m_text.find(strFind.data());
    textRange.length = strFind.length();
    DX::ThrowIfFailed(
        m_textLayout->SetFontStyle(DWRITE_FONT_STYLE_ITALIC, textRange)
        );

    strFind = L"IDWriteTextLayout";
    textRange.startPosition = m_text.find(strFind.data());
    textRange.length = strFind.length();
    DX::ThrowIfFailed(
        m_textLayout->SetFontStyle(DWRITE_FONT_STYLE_ITALIC, textRange)
        );

    strFind = L"SetDrawingEffect";
    textRange.startPosition = m_text.find(strFind.data());
    textRange.length = strFind.length();
    DX::ThrowIfFailed(
        m_textLayout->SetFontStyle(DWRITE_FONT_STYLE_ITALIC, textRange)
        );

    DX::ThrowIfFailed(
        m_textLayout->GetMetrics(&m_textMetrics)
        );

    // Instantiate BasicTextRenderer
    m_textRenderer = new BasicTextRenderer();

    CreateDeviceDependentResources();
}

void ColorTextFormattingDemoRenderer::CreateDeviceDependentResources()
{
    // Create brushes and set them
    ID2D1DeviceContext1* context = m_deviceResources->GetD2DDeviceContext();

    ComPtr<ID2D1SolidColorBrush> redBrush;
    DX::ThrowIfFailed(
        context->CreateSolidColorBrush(ColorF(ColorF::Red), &redBrush)
        );

    ComPtr<ID2D1SolidColorBrush> greenBrush;
    DX::ThrowIfFailed(
        context->CreateSolidColorBrush(ColorF(ColorF::Green), &greenBrush)
        );

    ComPtr<ID2D1SolidColorBrush> blueBrush;
    DX::ThrowIfFailed(
        context->CreateSolidColorBrush(ColorF(ColorF::Blue), &blueBrush)
        );

    DWRITE_TEXT_RANGE textRange;

    std::wstring strFind = L"RBG";
    textRange.startPosition = m_text.find(strFind.data());
    textRange.length = strFind.length();
    DX::ThrowIfFailed(
        m_textLayout->SetStrikethrough(true, textRange)
        );

    textRange.length = 1;
    DX::ThrowIfFailed(
        m_textLayout->SetDrawingEffect(redBrush.Get(), textRange)
        );

    textRange.startPosition += 1;
    DX::ThrowIfFailed(
        m_textLayout->SetDrawingEffect(blueBrush.Get(), textRange)
        );

    textRange.startPosition += 1;
    DX::ThrowIfFailed(
        m_textLayout->SetDrawingEffect(greenBrush.Get(), textRange)
        );

    strFind = L"RGB";
    textRange.startPosition = m_text.find(strFind.data());
    textRange.length = strFind.length();
    DX::ThrowIfFailed(
        m_textLayout->SetUnderline(true, textRange)
        );

    textRange.length = 1;
    DX::ThrowIfFailed(
        m_textLayout->SetDrawingEffect(redBrush.Get(), textRange)
        );

    textRange.startPosition += 1;
    DX::ThrowIfFailed(
        m_textLayout->SetDrawingEffect(greenBrush.Get(), textRange)
        );

    textRange.startPosition += 1;
    DX::ThrowIfFailed(
        m_textLayout->SetDrawingEffect(blueBrush.Get(), textRange)
        );

    strFind = L" red";      // avoid "rendered"
    textRange.startPosition = m_text.find(strFind.data()) + 1;
    textRange.length = strFind.length() - 1;
    DX::ThrowIfFailed(
        m_textLayout->SetDrawingEffect(redBrush.Get(), textRange)
        );

    strFind = L"green";
    textRange.startPosition = m_text.find(strFind.data());
    textRange.length = strFind.length();
    DX::ThrowIfFailed(
        m_textLayout->SetDrawingEffect(greenBrush.Get(), textRange)
        );

    strFind = L"blue";
    textRange.startPosition = m_text.find(strFind.data());
    textRange.length = strFind.length();
    DX::ThrowIfFailed(
        m_textLayout->SetDrawingEffect(blueBrush.Get(), textRange)
        );

    // Create default text brush
    DX::ThrowIfFailed(
        context->CreateSolidColorBrush(ColorF(ColorF::Black), &m_blackBrush)
        );
}
void ColorTextFormattingDemoRenderer::ReleaseDeviceDependentResources()
{
    m_blackBrush.Reset();
}

void ColorTextFormattingDemoRenderer::Update(DX::StepTimer const& timer)
{
}

// Renders a frame to the screen.
void ColorTextFormattingDemoRenderer::Render()
{
    ID2D1DeviceContext* context = m_deviceResources->GetD2DDeviceContext();
    Windows::Foundation::Size logicalSize = m_deviceResources->GetLogicalSize();

    context->SaveDrawingState(m_stateBlock.Get());
    context->BeginDraw();
    context->Clear(ColorF(ColorF::AliceBlue));

    // Center text on the screen
    Matrix3x2F screenTranslation = Matrix3x2F::Translation(
        (logicalSize.Width - m_textLayout->GetMaxWidth()) / 2,
        (logicalSize.Height - m_textMetrics.height) / 2);

    context->SetTransform(screenTranslation *
        m_deviceResources->GetOrientationTransform2D());

    // Display paragraph of text
    D2D1_POINT_2F origin = Point2F();

    if (!m_useCustomRenderer)
    {
        // Built-in text renderer
        context->DrawTextLayout(origin,
                                m_textLayout.Get(),
                                m_blackBrush.Get());
    }
    else
    {
        // Custom text renderer
        DrawingContext drawingContext(context,
                                      m_blackBrush.Get());

        m_textLayout->Draw(&drawingContext, m_textRenderer.Get(), 
                           origin.x, origin.y);
    }

    // Ignore D2DERR_RECREATE_TARGET here. This error indicates that the device
    // is lost. It will be handled during the next call to Present.
    HRESULT hr = context->EndDraw();
    if (hr != D2DERR_RECREATE_TARGET)
    {
        DX::ThrowIfFailed(hr);
    }

    context->RestoreDrawingState(m_stateBlock.Get());
}

void ColorTextFormattingDemoRenderer::SetUseCustomRenderer(bool use)
{
    m_useCustomRenderer = use;
}


